#include <Arduino.h>
#include <esp_task_wdt.h>
#include <algorithm>
#include <string>
#include <vector>
#include <WiFiClientSecure.h>

#include <Preferences.h>

#include <HCSR04.h>

#include <PubSubClient.h>

#include <AsyncTCP.h>
#include "ESPAsyncWebServer.h"

#include <HTTPClient.h>
#include <esp32fota.h>

AsyncWebServer server(80);

Preferences preferences;

WiFiClient espClient;
PubSubClient mqtt( espClient );

hw_timer_t *Timer0_Cfg = NULL;

UltraSonicDistanceSensor distanceCar( 18, 19 );
UltraSonicDistanceSensor distanceDoor( 4, 2 );
uint8_t relay[4] = { 14, 27, 16, 17 };

// Keep these variables during reboot
RTC_NOINIT_ATTR uint8_t car_state_save;
RTC_NOINIT_ATTR uint8_t door_state_save;

uint8_t car_state = 0;
uint8_t door_state = 0;
uint8_t block_door = 0;
uint8_t car_state_old = 1;
uint8_t door_state_old = 1;

double car_present_distance;
double door_open_distance;
double car_distance;
double door_distance;

long previousMillis;
uint8_t connect_retry_counter = 0;

// String vector pair of topic and message
std::vector< std::pair<std::string,std::string>> mqtt_msg;

// Timeout scheduler
typedef struct t {
    unsigned long tStart;
    unsigned long tTimeout;
} t;
t t_get_car_door = { 0, 10000 }; // poll inputs every 10 s
t t_chk_mqtt = { 0, 5000 }; // try to reconnect mqtt every 5 s
t t_mqtt_send = { 0, 1000 }; // send mqtt messages every s


esp32FOTA esp32FOTA( "esp-garage", 1, false, false );

bool tCheck( struct t *t ) {
  if( millis() > t->tStart + t->tTimeout ) return true;
  return false;
}

void tRun( struct t *t ) {
  t->tStart = millis();
}


void switch_relay( uint8_t n, uint8_t state ) {
    digitalWrite( relay[ n ], state );
}

void tap_switch( uint8_t n ) {
  switch_relay( n, HIGH );
  delay( 500 );
  switch_relay( n, LOW );
}

double get_distance( uint8_t sensor ) {
  double distance;
  // make 3 measurements and return the average
  for( uint8_t i = 0; i < 3; i++ ) {
    if( sensor == 0 ) {
      distance += distanceCar.measureDistanceCm( 0.0 );
    } else {
      distance += distanceDoor.measureDistanceCm( 0.0 );
    }
    delay( 50 );
  }
  return distance / 3; 
}

bool car_present() {
  double d = get_distance( 0 );
  if( d != -1 ) { // discard invalid measurements
    car_distance = d;
  }
  if( car_distance < car_present_distance ) {
    return true;
  }
  return false;
}

boolean door_open() {
  double d = get_distance( 1 );
  if( d != -1 ) { // discard invalid measurements
    door_distance = d;
  }
  if( door_distance < door_open_distance ) {
    return true;
  }
  return false;
}


void notFound(AsyncWebServerRequest *request) {
    request->send(404, "text/plain", "Not found\n");
}

void mqtt_publish( std::string topic = "", std::string message = "" ) {
  std::vector< std::pair<std::string,std::string>>::iterator elm;

  // if there's more than 20 messages queued, start removing the oldest ones
  if( mqtt_msg.size() > 20 ) {
    elm = mqtt_msg.begin();
    mqtt_msg.erase( elm );
  }

  if( topic != "" && message != "" ) {
    mqtt_msg.push_back( { topic, message } );
  }


  // only try sending for 500 ms at a time
  unsigned long currentMillis = millis();
  // refresh mqtt before checking connection status
  mqtt.loop();
  while( mqtt_msg.size() > 0 && millis() - currentMillis < 500 && mqtt.connected() ) {
    elm = mqtt_msg.begin();

    std::string topic = elm->first;
    std::string msg   = elm->second;

    if( mqtt.publish( ( "garage/" + elm->first ).c_str(), elm->second.c_str() ) ) {
      mqtt_msg.erase( elm );
    }
    esp_task_wdt_reset();
    mqtt.loop();
  }
}

void mqtt_connect() {
  if( mqtt.connect( "garage" ) ) {
    mqtt_publish( "$state", "READY" );
  } else {
    Serial.println( "Failed to connect to MQTT broker." );
  }
}

void IRAM_ATTR Timer0_ISR() {
  esp_task_wdt_reset();
}


void setup() {
  // put your setup code here, to run once:
  Serial.begin(115200);

  Serial.println( "Start" );

  preferences.begin( "garage", false );
  door_open_distance = preferences.getDouble( "door", 100.0 );
  car_present_distance = preferences.getDouble( "car", 100.0 );
  preferences.end();

  // If ESP was reset by SW or WDT load old values from nvram
  esp_reset_reason_t reason = esp_reset_reason();
  if( reason != ESP_RST_POWERON ) {
    car_state = car_state_save;
    car_state_old = car_state;

    door_state = door_state_save;
    door_state_old = door_state;
  }

  // initialize distances as door closed and car parked.
  door_distance = door_open_distance + 1;
  car_distance = car_present_distance - 1;

  for( uint8_t i = 0; i < 4; i++ ) {
    pinMode( relay[i], OUTPUT );
    digitalWrite( relay[i], LOW );
  }
  
  // Enable WDT 60s timeout
  esp_task_wdt_init( 60, true); //enable panic so ESP32 restarts
  esp_task_wdt_add(NULL); //add current thread to WDT watch
  // run an independant WDT reset
  Timer0_Cfg = timerBegin(0, 80, true);
  timerAttachInterrupt(Timer0_Cfg, &Timer0_ISR, true);
  timerAlarmWrite(Timer0_Cfg, 1000, true);
  timerAlarmEnable(Timer0_Cfg);

  WiFi.setHostname( "garage" );
  WiFi.mode( WIFI_STA );
  WiFi.begin( "SSID", "WIFI PASSWORD" );
  Serial.print( "Connecting to Wifi" );

  uint8_t timeout = 0;
  while( WiFi.status() != WL_CONNECTED && ++timeout < 600 ) {
    Serial.print(".");
    delay(1000);
  }
  Serial.println();
  if( timeout == 30 ) {
    Serial.println( "10m timeout reached, resetting." );
    delay( 3000 );
    ESP.restart();

  } else {
    Serial.println( "Connected!" );
    Serial.printf( "IP: %s\r\n", WiFi.localIP().toString().c_str() );
  }

  IPAddress mqtt_server;
  mqtt_server.fromString( "192.168.1.2" );
  Serial.println( "Set MQTT server." );
  mqtt.setServer( mqtt_server, 1883 );
  mqtt.setKeepAlive( 5 );
  mqtt.setSocketTimeout( 5 );

  server.on( "/get", HTTP_GET, [](AsyncWebServerRequest *request ) {
    double d_door = distanceDoor.measureDistanceCm( 0.0 );
    double d_car = distanceCar.measureDistanceCm( 0.0 );

    String reply = "{\"car\":\"" + (String)( car_state < 2 ? "parked" : "away" ) + "\",\"door\":\"" + (String)( door_state < 2 ? "closed" : "open" ) +
    ", \"sensors\": { \"car\": " + (String)d_car + ", \"door\": " + (String)d_door + " }" +
    ", \"limits\": { \"car\": " + (String) car_present_distance + ", \"door\": " + (String) door_open_distance + " }\n";

    request->send( 200, "javascript/json", reply );
  });

  server.on( "/tap", HTTP_GET,[](AsyncWebServerRequest *request ) {
    if( request->hasParam( "switch" ) ) {
      uint8_t sw = request->getParam( "switch" )->value().toInt();
      if( sw >= 0 && sw <= 3 ) {
        tap_switch( sw );
        request->send( 200, "text/plain", "OK\n" );
      } else {
        request->send( 500, "text/plain", "Bad request.\n" );
      }
    }
  });

  server.on( "/switch", HTTP_GET,[](AsyncWebServerRequest *request ) {
    if( request->hasParam( "switch" ) && request->hasParam( "state") ) {
      uint8_t sw = request->getParam( "switch" )->value().toInt();
      if( sw >= 0 && sw <= 3 ) {
        if( request->getParam( "state")->value().equals( "ON" ) ) {
          switch_relay( sw, HIGH );
        } else {
          switch_relay( sw, LOW );
        }
        request->send( 200, "text/plain", "OK\n" );
        return; 
      }
    }
    request->send( 500, "text/plain", "Bad request.\n" );
  });

  server.on( "/set", HTTP_GET, [](AsyncWebServerRequest *request ) {
    if( request->hasParam( "car" ) || request->hasParam( "door") ) {
      preferences.begin( "garage", true );
    } else {
      request->send( 500, "text/plain", "Bad request.\n" );
      return;
    }
    if( request->hasParam( "car" ) ) {
      double d = request->getParam( "car" )->value().toDouble();
      car_present_distance = d;
      preferences.putDouble( "car", d );
    }
    if( request->hasParam( "door" ) ) {
      double d = request->getParam( "door" )->value().toDouble();
      door_open_distance = d;
      preferences.putDouble( "door", d );
    }
    preferences.end();
    request->send( 200, "text/plain", "OK\n" );
  });

  server.on( "/update", HTTP_GET, [](AsyncWebServerRequest *request ) {
    if( request->hasParam( "address" ) && request->hasParam( "file" ) ) {
      esp32FOTA.forceUpdate( request->getParam( "address")->value().c_str(), 80, request->getParam( "file" )->value().c_str(), false );
    }
  });

  server.onNotFound(notFound);
  server.begin();

}

void loop() {
  // put your main code here, to run repeatedly:
  mqtt.loop();

  if( tCheck( &t_get_car_door ) ) {
    // if door blocker was set by returning car
    // decrease each time we're checking until
    // it is back at 0 again
    if( block_door > 0 ) {
      block_door--;
    }

    // let's check if the car is still here
    if( car_present() ) {
      if( car_state != 0 ) {
        Serial.println( "Car is back." );
        mqtt_publish( "car", "parked" );
      }
      car_state = 0;
    } else {
      if( car_state == 0 ) {
        Serial.println( "Car has left." );
        mqtt_publish( "car", "away" );
      }
      car_state++;
      // Only measuring 5 minutes
      if( car_state > 30 ) {
        car_state = 30;
      }
    }

    if( door_open() ) {
      if( door_state == 0 ) {
        Serial.println( "Door has opened." );
        mqtt_publish( "door", "open" );

        // If the door gets opened while the car is away
        // block the door for 5 minutes
        if( car_state >= 3 ) {
          block_door = 30;
        }
      }
      door_state++;

      // check door state for half hour
      if( door_state > 180 ) {
        door_state = 180;
      }
    } else {
      if( door_state != 0 ) {
        Serial.println( "Door has closed." );
        mqtt_publish( "door", "closed" );
        block_door = 0;
      }
      door_state = 0;
    }

    // If the car is gone for 30s and the door is open
    // close the door
    if( car_state == 3 && door_state >= 3 && block_door == 0 ) {
      Serial.println( "Car is gone for 1 minute, closing door." );
      mqtt_publish( "door", "closing" );
      door_state = 1;
      tap_switch( 3 );
    }

    // If the door is open for 30 minutes, close the door (doesn't matter what state car is)
    if( door_state == 180 ) {
      Serial.println( "Door is open for 30 minutes, closing door." );
      mqtt_publish( "door", "closing" );
      door_state = 1;
      tap_switch( 3 );
    }
    
    Serial.printf( "States: D: %d, C: %d\r\n", door_state, car_state );
    if( door_state != door_state_old ) {
      mqtt_publish( "door/$state", std::to_string( door_state ) );
      door_state_old = door_state;
    }
    if( car_state != car_state_old ) {
      mqtt_publish( "car/$state", std::to_string( car_state ) );
      car_state_old = car_state;
    }

    tRun( &t_get_car_door );
  }

  mqtt.loop();

  unsigned long currentMillis = millis();
  // if WiFi is down, try reconnecting
  if( WiFi.status() != WL_CONNECTED ) {
    if( currentMillis - previousMillis >= 10000 ) {
      if( connect_retry_counter > 60 ) {
        // failed to reconnect for 10 minutes, try a reset
        car_state_save = car_state;
        door_state_save = door_state;
        Serial.println( "Giving up on WiFi, restart..." );
        delay( 500 );
        ESP.restart();
      }
      Serial.print( millis() );
      Serial.println("Reconnecting to WiFi...");
      WiFi.disconnect();
      WiFi.reconnect();
      previousMillis = currentMillis;
      connect_retry_counter++;
    }
  } else {
    if( !mqtt.connected() ) {
      if( tCheck( &t_chk_mqtt) ) {
        mqtt_connect();
        tRun( &t_chk_mqtt );
      }
    } else {
      if( tCheck( &t_mqtt_send ) ) {
        if( mqtt_msg.size() > 0 ) {
          mqtt_publish();
        }
        tRun( &t_mqtt_send );
      }
    }
  }

  // save some energy.
  esp_task_wdt_reset();
  delay( 1 );
}
